
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mgl.h>

extern void mainMakeCheckImage (unsigned char *, int, int);
extern mglFont *mainFont;

#define MOUSE_PATTERNS 32

mglTexture *mouseTex [MOUSE_PATTERNS];


static int insertchar (char *str, int pos, char ch)
{
  int i;
  int len = strlen (str);

  if (pos <= len)
  {
    for (i = len+1; i > pos; i--)
      str [i] = str [i-1];
    str [pos] = ch;
    return 1;
  }
  return 0;
}

static int deletechar (char *str, int pos)
{
  int len = strlen (str);

  if ((pos < len) && (pos >= 0))
  {
    strcpy (str+pos, str+pos+1);
    return 1;
  }
  return 0;
}


int cursor [24] =
{
  GLUT_CURSOR_INHERIT,
  GLUT_CURSOR_NONE,
  GLUT_CURSOR_FULL_CROSSHAIR,
  GLUT_CURSOR_RIGHT_ARROW,
  GLUT_CURSOR_LEFT_ARROW,
  GLUT_CURSOR_INFO,
  GLUT_CURSOR_DESTROY,
  GLUT_CURSOR_HELP,
  GLUT_CURSOR_CYCLE,
  GLUT_CURSOR_SPRAY,
  GLUT_CURSOR_WAIT,
  GLUT_CURSOR_TEXT,
  GLUT_CURSOR_CROSSHAIR,
  GLUT_CURSOR_UP_DOWN,
  GLUT_CURSOR_LEFT_RIGHT,
  GLUT_CURSOR_TOP_SIDE,
  GLUT_CURSOR_BOTTOM_SIDE,
  GLUT_CURSOR_LEFT_SIDE,
  GLUT_CURSOR_RIGHT_SIDE,
  GLUT_CURSOR_TOP_LEFT_CORNER,
  GLUT_CURSOR_TOP_RIGHT_CORNER,
  GLUT_CURSOR_BOTTOM_RIGHT_CORNER,
  GLUT_CURSOR_BOTTOM_LEFT_CORNER
};


char *cursor_name [24] =
{
  "INHERIT",
  "NONE",
  "FULL_CROSSHAIR",
  "RIGHT_ARROW",
  "LEFT_ARROW",
  "INFO",
  "DESTROY",
  "HELP",
  "CYCLE",
  "SPRAY",
  "WAIT",
  "TEXT",
  "CROSSHAIR",
  "UP_DOWN",
  "LEFT_RIGHT",
  "TOP_SIDE",
  "BOTTOM_SIDE",
  "LEFT_SIDE",
  "RIGHT_SIDE",
  "TOP_LEFT_CORNER",
  "TOP_RIGHT_CORNER",
  "BOTTOM_RIGHT_CORNER",
  "BOTTOM_LEFT_CORNER",
  "User defined"
};

char *keyName;
int ascii;
int key;
char keyBuf [10];


char testStr [1000];
int testStrPos;
int testStrLen;

mglImage *chess;
mglTexture *chessTex;

mglTexture *key1;
mglTexture *key2;

void configInit (void)
{
  int i, j;
  mglImage *mouseimg, *img;
  unsigned char *dat, *pat;
  char filename[80];
  mglSetMouseCursor (GLUT_CURSOR_INHERIT);

  if (chess == NULL)
  {
    chess = mglCreateImage (32, 32, 1);
    mainMakeCheckImage (chess->data, 32, 16);
    chess->type = GL_LUMINANCE;
    chessTex = mglMakeTexture (chess, GL_NEAREST);

    key1 = mglLoadTexture ("../data/demo/left.tga", 0, GL_LINEAR);
    key2 = mglLoadTexture ("../data/demo/right.tga", 0, GL_LINEAR);

    mouseimg = mglLoadImage ("../data/mouse/mouse.tga", 2);

    for (i=0; i < MOUSE_PATTERNS; i++)
    {
      sprintf(filename, "../data/mouse/caust%02d.bw", i);
      pat = (unsigned char *) mouseimg->data;
      img = mglLoadImage (filename, 2);
      dat = (unsigned char *) img->data;
      for (j = 0; j < img->w * img->h; j++)
      {
        if ((*pat) < 127) *dat = 0;
        pat++;
        dat++;
        if (img->d == 2)
        {
          if (!(*pat) < 127) *dat = 0;
          {
            dat++;
            pat++;
          }
        }
      }
      mouseTex [i] = mglMakeTexture (img, GL_NEAREST_MIPMAP_NEAREST);
      mglDeleteImage (img);
    }
    mglDeleteImage (mouseimg);
    mglDeleteImage (chess);
  }

  testStr [0] = 0;
  testStrPos = 0;
  ascii = 0;
  keyName = "None";
}

void semafor (char *name, int ison, int x, int y, int sizeX, int sizeY)
{
  int posx;
  int posy;
  mglColor c;
  int len;
  int sx;
  int sy;

  len = strlen (name);

  posx = (sizeX - len * mainFont->w) / 2;
  posy = (sizeY - mainFont->h) / 2;
  sx = x + sizeX / 2;
  sy = y + sizeY / 2;

  if (ison)
  {
    c.r = 0;
    c.g = 0.5;
    c.b = 0;
    c.a = 0.05f;
    mglTriangleColor [0] = mglWhite;
    mglTriangleColor [1] = mglTriangleColor [2] = c;
    mglDrawFilledTriangle (sx, sy, x, y, x + sizeX, y);
    mglDrawFilledTriangle (sx, sy, x + sizeX, y, x + sizeX, y + sizeY);
    mglDrawFilledTriangle (sx, sy, x + sizeX, y + sizeY, x, y + sizeY);
    mglDrawFilledTriangle (sx, sy, x, y + sizeY, x, y);

    mglSetRectangleColors (c);
    mglDrawFilledRectangle (x, y, sizeX, sizeY);

    c.r = 0.3f;
    c.g = 1;
    c.b = 1;
    c.a = 1;
    mglSetTextureColors (c);
    mglPuts (mainFont, name, x+posx, y+posy);
  }
  else
  {
    c.r = 0.333f;
    c.g = 0;
    c.b = 0;
    c.a = 1;
    mglTriangleColor [0] = c;
    c = mglBlack;
    c.a = 0.05f;
    mglTriangleColor [1] = mglTriangleColor [2] = c;
    mglDrawFilledTriangle (sx, sy, x, y, x + sizeX, y);
    mglDrawFilledTriangle (sx, sy, x + sizeX, y, x + sizeX, y + sizeY);
    mglDrawFilledTriangle (sx, sy, x + sizeX, y + sizeY, x, y + sizeY);
    mglDrawFilledTriangle (sx, sy, x, y + sizeY, x, y);

    c.r = 0.1f;
    c.g = 0.6f;
    c.b = 0.6f;
    c.a = 0.333f;
    mglSetTextureColors (c);
    mglPuts (mainFont, name, x+posx, y+posy);
  }
}




void configDisplay (void)
{
  int posx;
  int posy;
  char s [100];
  int mx1, mx2, my1, my2;
  int cx, cy;
  int tax, tay;
  mglColor c;
  mglColor textc;

  c.r = 0.6f;
  c.g = 0.8f;
  c.b = 1.0f;
  c.a = 1;

  textc = mglWhite;

  c.r = 0.5;
  c.g = 0.5;
  c.b = 0.5;
  c.a = 1;
  mglSetTextureColors (c);
  mglSetTextureZoom (1);
  mglDrawTexturedRectangle (key1, 0, 0, 400, 600);
  mglDrawTexturedRectangle (key2, 399, 0, 400, 600);

  posx = 2*mainFont->w;
  posy = 50;

  mglSetTextureColors (mglBlue);
  mglSetTextureZoom (1.5);
  mglPuts (mainFont, "Keyboard test", posx, posy);
  mglPuts (mainFont, "Mouse test", 400-mainFont->w + posx, posy);
  mglSetTextureZoom (1);

  posy += 2*mainFont->h + 20;
  mglSetTextureColors (textc);
  mglPuts (mainFont, "Modifiers", posx, posy);
  mglPuts (mainFont, "Mouse buttons", 400-mainFont->w + posx, posy);

  posy += mainFont->h;
  semafor ("Ctrl", mglCtrl, posx, posy, 6*mainFont->w, 2*mainFont->h);
  semafor ("Alt", mglAlt, posx + 8*mainFont->w, posy, 5*mainFont->w, 2*mainFont->h);
  semafor ("Shift", mglShift, posx + 15*mainFont->w, posy, 7*mainFont->w, 2*mainFont->h);

  semafor ("Left", mglMouseLeft, 400-mainFont->w+posx, posy, 6*mainFont->w, 2*mainFont->h);
  semafor ("Mid", mglMouseMiddle, 400-mainFont->w+posx + 8*mainFont->w, posy, 5*mainFont->w, 2*mainFont->h);
  semafor ("Right", mglMouseRight, 400-mainFont->w+posx + 15*mainFont->w, posy, 7*mainFont->w, 2*mainFont->h);

  posy += 3*mainFont->h;

  mglSetTextureColors (textc);
  mglPuts (mainFont, "Value   Key", posx, posy);
  mglPuts (mainFont, "Mouse position", 400-mainFont->w + posx, posy);
  posy += mainFont->h;
  sprintf (s, "%i", ascii);
  semafor (s, (ascii > 0) && (ascii < 256), posx, posy, 6*mainFont->w, 2*mainFont->h);
  semafor (keyName, strcmp (keyName, "None"), posx + 8*mainFont->w, posy, 5*mainFont->w, 2*mainFont->h);

  sprintf (s, "X=%.0f", mglMouseX);
  semafor (s, mglMouseIn, 400-mainFont->w + posx, posy, 8*mainFont->w, 2*mainFont->h);
  sprintf (s, "Y=%.0f", mglMouseY);
  semafor (s, mglMouseIn, 400-mainFont->w + posx + 11*mainFont->w, posy, 7*mainFont->w, 2*mainFont->h);

  posy += 3 * mainFont->h;

  mx1 = 400-mainFont->w+posx;
  my1 = posy + mainFont->h;
  mx2 = 800-2* mainFont->w;
  my2 = 600- 2 * mainFont->h;

  if ((mglMouseX >= mx1 && mglMouseX <= mx2) && (mglMouseY >= my1 && mglMouseY <= my2))
  {
    cx = mglMouseX - mx1;
    cx = (6 * cx) / (mx2 - mx1);

    cy = mglMouseY - my1;
    cy = (4 * cy) / (my2 - my1);

    cx += 6 * cy;
  }
  else
    cx = -1;

  mglSetTextureColors (textc);
  mglPuts (mainFont, "Test area", posx, posy);
  if ((cx == -1) || (cursor_name [cx] == NULL) || cx > 23)
    sprintf (s, "%s", "Pick mouse cursor here");
  else
    sprintf (s, "%s", cursor_name [cx]);

  mglPuts (mainFont, s, 400-mainFont->w + posx, posy);

  posy += mainFont->h;

  c = mglBlack;
  c.a = 0.5;
  mglSetRectangleColors (c);

  tax = 400-mainFont->w - posx;
  tay = 600 - posy - 2 * mainFont->h;
  mglDrawFilledRectangle (posx, posy, tax, tay);

  mglTextureZoomX = 0.333f;
  mglTextureZoomY = 0.5f;

  c = mglWhite;
  c.a = 0.5;
  mglSetTextureColors (c);
  mglDrawTexturedRectangle (chessTex, mx1, my1, mx2 - mx1, my2 - my1);
  mglSetTextureZoom (1);

  if (mglSecs (mglTime) % 2)
  {
    insertchar (testStr, testStrPos, '_');
  }
  mglSetTextureColors (mglWhite);
  mglFontWrap = tax / mainFont->w - 1;
  testStrLen = mglFontWrap * ((tay -3) / mainFont->h) - 3;
  mglFontWordWrap = 0;
  mglPuts (mainFont, testStr, posx + mainFont->w/2, posy + mainFont->h/2);
  mglFontWrap = 0;
  if (mglSecs (mglTime) % 2)
  {
    deletechar (testStr, testStrPos);
  }

  if ((cx == -1) || (cx == 23) || (!cursor_name [cx]))
  {
    if (mglMouseIn)
    {
      // update mouse cursor
      mglSetMouseCursor (GLUT_CURSOR_NONE);
      mglSetTextureZoom (0.889);
      mglDrawTexture (mouseTex [mglTime/1000 % MOUSE_PATTERNS], mglMouseX, mglMouseY);
      mglSetTextureZoom (1.0);
    }
  }
  else
    if (cursor [cx])
    {
      mglSetMouseCursor (cursor [cx]);
    }
}


void configKeyboard (void)
{
  ascii = mglLastKey;

  switch (mglLastKey)
  {
    case MGL_KEY_F1:
      keyName = "F1";
      break;
    case MGL_KEY_F2:
      keyName = "F2";
      break;
    case MGL_KEY_F3:
      keyName = "F3";
      break;
    case MGL_KEY_F4:
      keyName = "F4";
      break;
    case MGL_KEY_F5:
      keyName = "F5";
      break;
    case MGL_KEY_F6:
      keyName = "F6";
      break;
    case MGL_KEY_F7:
      keyName = "F7";
      break;
    case MGL_KEY_F8:
      keyName = "F8";
      break;
    case MGL_KEY_F9:
      keyName = "F9";
      break;
    case MGL_KEY_F10:
      keyName = "F10";
      break;
    case MGL_KEY_F11:
      keyName = "F11";
      break;
    case MGL_KEY_F12:
      keyName = "F12";
      break;
    case MGL_KEY_INSERT:
      keyName = "Ins";
      break;
    case MGL_KEY_HOME:
      keyName = "Home";
      break;
    case MGL_KEY_END:
      keyName = "End";
      break;
    case MGL_KEY_PAGE_UP:
      keyName = "PgUp";
      break;
    case MGL_KEY_PAGE_DOWN:
      keyName = "PgDn";
      break;
    case MGL_KEY_LEFT:
      keyName = "Left";
      break;
    case MGL_KEY_RIGHT:
      keyName = "Right";
      break;
    case MGL_KEY_UP:
      keyName = "Up";
      break;
    case MGL_KEY_DOWN:
      keyName = "Down";
      break;
    case '\b':
      if (deletechar (testStr, testStrPos-1))
        testStrPos--;
      keyName = "<-";
      break;
    case '\r':
    case '\n':
      keyName = "Enter";
      testStr [0] = 0;
      testStrPos = 0;
      break;
    case '\t':
    case '\v':
      keyName = "Tab";
      break;
    case 127:
      keyName = "Del";
      break;
    case 27:
      keyName = "Esc";
      break;
    default:
      keyName = "None";
      if (mglLastKey < 256)
      {
        sprintf (keyBuf, "\'%c\'", ascii);
        keyName = keyBuf;
        if (testStrPos <= testStrLen)
        {
          insertchar (testStr, testStrPos, (char)ascii);
          testStrPos++;
        }
      }
  }
}
