
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mgl.h>

extern int mainOption;
extern mglFont *mainFont;
extern int mainPaused;
extern int mainSpeed;

#define POINT 0
#define LINE 1
#define TRIANGLE 2
#define FILLED_TRIANGLE 3
#define RECTANGLE 4
#define FILLED_RECTANGLE 5
#define TEXTURE 6
#define SUBTEXTURE 7
#define TEXTURED_RECTANGLE 8
#define STRING 9

#define DIRNAME "../data/tests/"

#define T8BMP "test8.bmp"
#define T8TGA "test8.tga"
#define T8SGI "test8.bw"

#define T24BMP "test24.bmp"
#define T24TGA "test24.tga"
#define T24SGI "test24.rgb"

#define T32TGA "test32.tga"
#define T32SGI "test32.rgb"

#define TFNTBMP "testfont.tga"
#define TFNTTXT "testfont.txt"

#define TTXT "teststr.txt"

static char *imagesToLoad []   = {T24SGI, T8BMP, T8TGA, T8SGI, T24BMP, T24TGA, T24SGI, T32TGA, T32SGI};
static mglTexture *tex [] =     { NULL,  NULL,  NULL,   NULL,   NULL,   NULL,  NULL,   NULL, NULL};
static int texTotal;

static int loadStrings;
static int stringLines;
static int stringCharacters;
static char strings [80] [1000];

mglFont *testFont;

/*
#define POINT_MAX 30
#define LINE_MAX 10
#define TRIANGLE_MAX 5
#define RECTANGLE_MAX 5
*/

//static int objCounts [LAST_OBJECT_TYPE+1] = {30, 10, 5, 5, 7, 7, 8, 16};

typedef struct
{
  int type;
  float x [3];
  float y [3];
  int len;
  int dirx;
  int diry;
  float zoomX;
  float zoomY;
  int subX;
  int subY;
  int subW;
  int subH;
  mglColor c [4];
  float rotation;

  mglTexture *tex;
  char *text;
  mglFont *font;
} Tobj;

static Tobj *objs [1000];
static int objsMax;


static int testsOption=-1;
static int shading;
static int alpha;
static int testMode;
static int objectsMode;
static int loaded;
static char loadStr [1000];
static int loadState;
static int loadFont;

float randf (float number)
{
	int r;

	r = rand () % 1001;
	number *= r / 1000.0;
	return number;
}

static Tobj *newObject (int type)
{
  Tobj *obj;
  int s, t;
  int i;
  int pointsize;

  obj = _calloc (sizeof (Tobj), 1);
  obj->type = type;

  s = randf (mglVirtualWidth / 2);
  t = randf (mglVirtualWidth / 2);
  if (s < t)
  {
    obj->x [0] = s;
    obj->x [1] = t;
  }
  else
  {
    obj->x [0] = t;
    obj->x [1] = s;
  }

  s = randf (mglVirtualHeight / 2);
  t = randf (mglVirtualHeight / 2);
  if (s < t)
  {
    obj->y [0] = s;
    obj->y [1] = t;
  }
  else
  {
    obj->y [0] = t;
    obj->y [1] = s;
  }

  obj->x [2] = randf (mglVirtualWidth / 2);
  obj->y [2] = randf (mglVirtualHeight / 2);

  obj->rotation = randf (359);

  while (!obj->dirx)
    obj->dirx = 5 - randf (10);
  while (!obj->diry)
    obj->diry = 5 - randf (10);


  if ((pointsize = mglMaxPointSize) > 100)
	  pointsize = 100;

  obj->zoomX = 1 + randf (1) * (pointsize-1);
  obj->zoomY = 1 + randf (1) * (mglMaxLineSize-1);

  switch (type)
  {
    case STRING:
      obj->len = 1;
      obj->zoomX = 0.5 + randf (599) / 333;
      obj->zoomY = 0.5 + randf (599) / 333;
       break;
    case LINE:
      obj->len = 2;
      break;
    case TRIANGLE:
    case FILLED_TRIANGLE:
      obj->len = 3;
      break;
    case TEXTURED_RECTANGLE:
      obj->zoomX = 2 - randf (1799) / 999;
      obj->zoomY = 2 - randf (1799) / 999;
      break;
    default:
      obj->len = 4;
      break;
  }

  for (i = 0; i < 4; i++)
  {
    obj->c [i].r = randf (1);
    obj->c [i].g = randf (1);
    obj->c [i].b = randf (1);
    if (obj->type == STRING)
      obj->c [i].a = 1;
    else
      obj->c [i].a = randf (1);
  }

  return obj;
}

static void deleteObject (Tobj *obj)
{
  if (!obj)
    return;

//  mglDeleteTexture (obj->tex);

  _free (obj);
}

static void deleteAllObjects (void)
{
  int i;

  for (i = 0; i < objsMax; i++)
    deleteObject (objs [i]);

  objsMax = 0;
}


static float p (float x)
{
  if (mglFps)
  {
    x *= (((float)mainSpeed / 5) * 100) / mglFps;
  }
  if (x > 99)
    return 99;

  return x;
}


static void moveObject (Tobj *obj)
{
  int i;
  int xp, yp;

  for (i = 0; i < 2; i++)
  {
    xp = obj->x [i] + p (obj->dirx);
    yp = obj->y [i] + p (obj->diry);

    if ((xp < 0) || (xp > mglVirtualWidth))
      obj->dirx = -obj->dirx;
    if ((yp < 0) || (yp > mglVirtualHeight))
      obj->diry = -obj->diry;
  }

  for (i = 0; i < 2; i++)
  {
    obj->x [i] += p (obj->dirx);
    obj->y [i] += p (obj->diry);
  }

  if (mainSpeed)
  {
    if (mglFps)
    {
      obj->rotation += (float)30 / mglFps;
      if (obj->rotation > 360)
        obj->rotation -= 360;
    }
  }
}

static void displayObject (Tobj *obj)
{
  mglRotationAngle = obj->rotation;

  mglRectangleColor [0] = obj->c [0];
  mglRectangleColor [1] = obj->c [1];
  mglRectangleColor [2] = obj->c [2];
  mglRectangleColor [3] = obj->c [3];

  mglTriangleColor [0] = obj->c [0];
  mglTriangleColor [1] = obj->c [1];
  mglTriangleColor [2] = obj->c [2];

  if ((shading) || (obj->type == STRING))
  {
    mglTextureColor [0] = obj->c [0];
    mglTextureColor [1] = obj->c [1];
    mglTextureColor [2] = obj->c [2];
    mglTextureColor [3] = obj->c [3];
  }
  else
  {
    //mglSetTextureColors (mglWhite);
  }

  switch (obj->type)
  {
    case POINT:
      mglPointColor = (obj->c [0]);
      mglSetPointSize (obj->zoomX);
      mglDrawPoint (obj->x [0], obj->y [0]);
      break;
    case LINE:
      mglLineColor [0] = obj->c [0];
      mglLineColor [1] = obj->c [1];
      mglSetLineSize (obj->zoomY);
      mglDrawLine (obj->x [0], obj->y [0], obj->x [1], obj->y [1]);
      break;
    case TRIANGLE:
      mglSetLineSize (obj->zoomY);
      mglDrawTriangle (obj->x [0], obj->y [0], obj->x [1], obj->y [1], obj->x [2], obj->y [2]);
      break;
    case RECTANGLE:
      mglSetLineSize (obj->zoomY);
      mglDrawRectangle (obj->x [0], obj->y [0], obj->x [1] - obj->x [0], obj->y [1] - obj->y [0]);
      break;
    case FILLED_TRIANGLE:
      mglDrawFilledTriangle (obj->x [0], obj->y [0], obj->x [1], obj->y [1], obj->x [2], obj->y [2]);
      break;
    case FILLED_RECTANGLE:
      mglDrawFilledRectangle (obj->x [0], obj->y [0], obj->x [1] - obj->x [0], obj->y [1] - obj->y [0]);
      break;
    case TEXTURE:
      mglDrawTexture (obj->tex, obj->x [0], obj->y [0]);
      break;
    case SUBTEXTURE:
      mglDrawSubTexture (obj->tex, obj->x [0], obj->y [0], obj->subX, obj->subY, obj->subW, obj->subH);
      break;
    case TEXTURED_RECTANGLE:
      mglTextureZoomX = obj->zoomX;
      mglTextureZoomY = obj->zoomY;
      mglDrawTexturedRectangle (obj->tex, obj->x [0], obj->y [0], obj->subX, obj->subY);
      mglTextureZoomX = 1;
      mglTextureZoomY = 1;
      break;
    case STRING:
      mglTextureZoomX = obj->zoomX;
      mglTextureZoomY = obj->zoomY;
      mglFontWordWrap = 1;
      mglFontWrap = obj->subW;
      mglPuts (testFont, obj->text, obj->x [0], obj->y [0]);
      mglFontWrap = 0;
      mglFontWordWrap = 0;
      mglTextureZoomX = 1;
      mglTextureZoomY = 1;
      break;
    default:
      break;
  }
}



static void addObjects (int type, int count)
{
  int i;
  Tobj *obj;
  int len;
  int w;
  int h;

  for (i = 0; i < count; i++)
  {
    obj = objs [objsMax] = newObject (type);
    objsMax++;

    switch (obj->type)
    {
      case TEXTURE:
        obj->tex = tex [i];
        break;
      case TEXTURED_RECTANGLE:
        obj->tex = tex [i];
        w = obj->tex->image.w;
        h = obj->tex->image.h;
        obj->subX = w/2 + rand () % (int) (w * 1.5f);
        obj->subY = h/2 + rand () % (int) (h * 1.5f);
        break;
      case SUBTEXTURE:
        obj->tex = tex [i];
        w = obj->tex->image.w;
        h = obj->tex->image.h;
        obj->subW = w/3 + rand () % (w/2);
        obj->subH = h/3 + rand () % (h/2);
        obj->subX = rand () % (w - obj->subW);
        obj->subY = rand () % (h - obj->subH);
        break;
      case STRING:
        obj->text = strings [i];
        len = strlen (strings [i]);
        obj->subW = 1000 / (2 * mainFont->w * obj->zoomX);
        obj->font = mainFont;
        break;
    }
  }
}


static void displayObjects (void)
{
  int i;

  for (i = 0; i < objsMax; i++)
  {
    if (!mainPaused)
      moveObject (objs [i]);

    displayObject (objs [i]);
  }
}


static void addAllObjects (void)
{
  int i;

  if (testMode == objectsMode)
    return;

  objectsMode = testMode;

  deleteAllObjects ();

  switch (objectsMode)
  {
    case 0:
      addObjects (POINT, 20);
      addObjects (LINE, 7);
      addObjects (TRIANGLE, 3);
      addObjects (RECTANGLE, 3);
      break;
    case 1:
      addObjects (FILLED_TRIANGLE, 5);
      addObjects (FILLED_RECTANGLE, 5);
      break;
    case 2:
      addObjects (TEXTURE, texTotal);
      break;
    case 3:
      for (i = 0; i < 2; i++)
      {
        addObjects (SUBTEXTURE, texTotal);
      }
      break;
    case 4:
      addObjects (STRING, stringLines);
      break;
    case 5:
      addObjects (TEXTURED_RECTANGLE, texTotal);
      break;
    default:
      break;
  }
}


static void setTestsOption (void)
{
  if (testsOption == mainOption)
    return;
  testsOption = mainOption;

  switch (testsOption)
  {
    case 1:
      addAllObjects ();
      shading = 0;
      alpha = 0;
      mglSetWindowTitle ("Grafick primitiva knihovny MGL - bez interpolace barev a bez pruhlednosti");
      break;
    case 2:
      addAllObjects ();
      shading = 0;
      alpha = 1;
      mglSetWindowTitle ("Grafick primitiva knihovny MGL - bez interpolace barev s pruhlednost");
      break;
    case 3:
      addAllObjects ();
      shading = 1;
      alpha = 0;
      mglSetWindowTitle ("Grafick primitiva knihovny MGL - s interpolac barev bez pruhlednosti");
      break;
    case 4:
      addAllObjects ();
      shading = 1;
      alpha = 1;
      mglSetWindowTitle ("Grafick primitiva knihovny MGL - s interpolac barev a s pruhlednost");
      break;
    default:
      mainOption = 1;
      setTestsOption ();
      break;
   }
}

void testsInit (void)
{
  //mainOption = 0;
  //testsOption = -1;
  testMode = -1;
  loadState = -1;
}


void testsDisplay (int mode)
{
  mglColor c;

  if (testMode != mode)
  {
    testMode = mode;
    objectsMode = -1;
    testsOption = -1;
  }

  c = mglCeleste;
  mglClearScreen (c);

  if (shading)
    mglInterpolationOn ();
  else
    mglInterpolationOff ();

  if (alpha)
    mglAlphaOn ();
  else
    mglAlphaOff ();

  setTestsOption ();

  displayObjects ();

  mglRotationAngle = 0;

  mglAlphaOn ();
  mglInterpolationOn ();
}

void testsDisplayFilled (void)
{
  testsDisplay (1);
}

void testsDisplayPrimitives (void)
{
  testsDisplay (0);
}


static int enter2continue;

void testsKeyboard (void)
{
  switch (mglLastKey)
  {
    case '\r':
      enter2continue = 0;
      break;
    default:
      break;
  }
}

static void readStrings (char *textFileName)
{
  FILE *f;
  char s [80];

  if ( ( f = fopen( textFileName, "rt" ) ) == NULL ) {
    sprintf (s, "Error opening string text file (%s)", textFileName);
		mglMessage ("Error", s);
    exit (1);
	}

  stringLines = 0;
  stringCharacters = 0;
  while (fgets (strings [stringLines], 80, f) != NULL)
  {
    if (strings [stringLines] [0] == '\n')
      continue;
    stringCharacters += strlen (strings [stringLines]) - 1;
    stringLines ++;
  }

  if (stringLines == 0)
  {
    sprintf (s, "Text file contains no text (%s) !", textFileName);
		mglMessage ("Error", s);
    exit (1);
  }

  fclose (f);
}

int enterTime;

void testsDisplayTexturesEngine (int mode)
{
  char fileName [333];
  int imagesToLoadTotal = (sizeof (imagesToLoad) / sizeof (char *));
  char s [100];
  mglTexture *t;
  mglColor c;

  if ((mglTime - enterTime) >= 5000)
  {
    enter2continue = 0;
  }

  if (!loaded)
  {
    if (loadState == -1)
    {
      strcpy (loadStr, "\n\a6Loading all neccessary data for testing\a0\n\n");
      loadState = 0;
      sprintf (s, "Loading texture %s ... ", imagesToLoad [loadState]);
      strcat (loadStr, s);
    }
    else
    {
      if (loadState < imagesToLoadTotal)
      {
        strcpy (fileName, DIRNAME);
        strcat (fileName, imagesToLoad [loadState]);
        if (!tex [loadState])
        {
          tex [loadState] = mglLoadTexture (fileName, 0, GL_NEAREST_MIPMAP_NEAREST);
          mglDeleteTexture (tex [loadState]);
          tex [loadState] = mglLoadTexture (fileName, 1, GL_NEAREST);
          mglDeleteTexture (tex [loadState]);
          tex [loadState] = mglLoadTexture (fileName, 2, GL_LINEAR);
        }
        t = tex [loadState];
        sprintf (s, "\a3OK\a5 (%i x %i x %i bpp)\a0\n", t->image.w, t->image.h, 8*t->image.d);
        strcat (loadStr, s);
        loadState++;
        if (loadState < imagesToLoadTotal)
        {
          strcpy (fileName, DIRNAME);
          strcat (fileName, imagesToLoad [loadState]);
          sprintf (s, "Loading texture %s ... ", imagesToLoad [loadState]);
          strcat (loadStr, s);
        }
        else
        {
          sprintf (s, "Loading font "TFNTBMP","TFNTTXT" ... ");
          strcat (loadStr, s);
          loadFont = 1;
        }
      }
      else
      {
        if (loadFont)
        {
          loadFont = 0;
          loadStrings = 1;
          if (!testFont)
            testFont = mglLoadFont (DIRNAME""TFNTTXT, DIRNAME""TFNTBMP);

          sprintf (s, "\a3OK\a5 (%i x %i)\a0\n", testFont->w, testFont->h);
          strcat (loadStr, s);
          sprintf (s, "Loading strings "TTXT" ... ");
          strcat (loadStr, s);
        }
        if (loadStrings)
        {
          if (stringLines == 0)
          {
            readStrings (DIRNAME""TTXT"");
            loaded = 1;
            texTotal = loadState;
            enter2continue = 1;
            enterTime = mglTime;
            sprintf (s, "\a3OK\a5 (%i lines, %i characters)\a0\n\nPlease wait ...", stringLines, stringCharacters);
            strcat (loadStr, s);
            loadStrings = 0;
          }
        }
      }
    }
  }
  else
  {
    if (!enter2continue && loaded)
    {
      testsDisplay (mode);
    }
  }

  if (!loaded || enter2continue)
  {
    mglClearScreen (mglBlack);
    mglSetTextureColors (mglWhite);

    c = mglBlack;
    c.a = 0.7f;
    mglSetRectangleColors (c);
    mglDrawFilledRectangle (24, 24, mglVirtualWidth - 48, mglVirtualHeight - 48);

    mglFontWrap = 0;
    mglSetTextureZoom (0.8f);
    mglPuts (mainFont, loadStr, mainFont->w, mainFont->h);
  }
}


void testsDisplayTextures (void)
{
  testsDisplayTexturesEngine (2);
}


void testsDisplaySubTextures (void)
{
  testsDisplayTexturesEngine (3);
}

void testsDisplayFonts (void)
{
  testsDisplayTexturesEngine (4);
}

void testsDisplayTexturedRectangles (void)
{
  testsDisplayTexturesEngine (5);
}
