
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <mgl.h>
#include <string.h>

mglColor mglBlack = {0, 0, 0, 1};
mglColor mglWhite = {1, 1, 1, 1};
mglColor mglRed = {1, 0, 0, 1};
mglColor mglGreen = {0, 1, 0, 1};
mglColor mglBlue = {0, 0, 1, 1};
mglColor mglYellow = {1, 1, 0, 1};
mglColor mglCyan = {0, 1, 1, 1};
mglColor mglMagenta = {1, 0, 1, 1};
mglColor mglBlank = {0, 0, 0, 0};

mglColor mglPink = {1, 0.5f, 1, 1};
mglColor mglCeleste = {0.6f, 0.8f, 1, 1};

mglColor *mglColorBase [] = {&mglWhite, &mglRed, &mglGreen, &mglBlue, &mglYellow,
&mglCyan, &mglMagenta, &mglBlack, &mglBlank, &mglPink, &mglCeleste};

mglColor mglRectangleColor [4];
mglColor mglFontColor [4];
mglColor mglTriangleColor [3];
mglColor mglPointColor;
mglColor mglLineColor [2];

int mglColorMode;

float mglMaxPointSize;
float mglMaxLineSize;

float mglRotationAngle;


static float coord (float x, float min, float max)
{
	return min + x * (max - min);
}

mglColor mglMixColors (mglColor color1, mglColor color2, float rate)
{
	mglColor col;

	col.r = coord (rate, color1.r, color2.r);
	col.g = coord (rate, color1.g, color2.g);
	col.b = coord (rate, color1.b, color2.b);
	col.a = coord (rate, color1.a, color2.a);

	return col;
}

void mglGetSizeRanges (void)
{
  GLfloat ranges [10];

  glGetFloatv (GL_POINT_SIZE_RANGE, ranges);
  mglMaxPointSize = ranges [1];

  glGetFloatv (GL_LINE_WIDTH_RANGE, ranges);
  mglMaxLineSize = ranges [1];

  mglTestOpenGLError ();
}

int mglFullScreenState;

int mglWindowPosX;
int mglWindowPosY;
int mglWindowSizeX;
int mglWindowSizeY;


static int mglSmoothState;
static int mglPointSmoothState;
static int mglAlphaState;
static int mglLineStippleState;
static int mglShadeState;
static int mglModeState;
static int mglBeginEndState;
static int mglFillState;
static int mglTextureEnableState;
static float mglLineState;
static float mglPointState;
static mglColor mglColorState;
extern int mglLastUsedTexture;

static void _mglAbort_ (void)
{
  if (mglBeginEndState == 1)
  {
    glEnd ();
    mglTestOpenGLError ();
    mglBeginEndState = 0;
/*
if (mglModeState == GL_POINTS)
{         if (mglAlphaState)
	   glEnable (GL_BLEND);
	 else
	   glDisable (GL_BLEND);
}*/
//    mglModeState = -1;
  }
  mglModeState = -1;
}


static void _mglBegin_ (int mode)
{
  if (mglModeState != mode)
  {
    _mglAbort_ ();

    if ((mglTextureEnableState == 1) && (mode == GL_QUADS))
		mglAlphaState ? glEnable (GL_BLEND) : glDisable (GL_BLEND);

    mglTestOpenGLError ();
    glBegin (mode);

    mglBeginEndState = 1;
    mglModeState = mode;
  }
}


static void _mglEnd_ (void)
{

}


static void _mglFillOn_ (void)
{
  if (mglFillState == 0)
  {
    _mglAbort_ ();
    glPolygonMode (GL_FRONT, GL_FILL);
    mglFillState = 1;
  }
}

static void _mglFillOff_ (void)
{
  if (mglFillState == 1)
  {
    _mglAbort_ ();
    glPolygonMode (GL_FRONT, GL_LINE);
    mglFillState = 0;
  }
}

mglColor mglAdjustColor (mglColor c, int mode)
{
	if (mode & MGL_COLORMODE_INVERT)
	{
		c.r = 1 - c.r; c.g = 1 - c.g; c.b = 1 - c.b;
	}
	if (mode & MGL_COLORMODE_SHADES)
	{
		c.r = c.g = c.b = (c.r + c.g + c.b / 3);
	}
	return c;
}

void mglSetColor (mglColor color)
{
	if (color.r != -1 && color.g != -1 && color.b != -1 && color.a != -1)
	{
		if (mglColorMode) color = mglAdjustColor (color, mglColorMode);

		if ((mglColorState.r != color.r) || (mglColorState.g != color.g) ||
			(mglColorState.b != color.b) || (mglColorState.a != color.a))
			{
				glColor4f (color.r, color.g, color.b, color.a);
				
			}
			return;
	}
	mglColorState = color;
}

void mglDoneState (void)
{
  _mglAbort_ ();
}


void mglInitState (void)
{
	mglColor c;

  mglDoneState ();

  glDisable (GL_LINE_SMOOTH);
  glDisable (GL_POINT_SMOOTH);
  glDisable (GL_POLYGON_SMOOTH);

  mglSmoothState = mglPointSmoothState = 0;

//  glEnable(GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//  glEnable(GL_ALPHA_TEST);
//  glAlphaFunc(GL_GEQUAL, 0.0005f);
  mglAlphaState = 1;
  glEnable (GL_BLEND);

  glShadeModel (GL_SMOOTH);
  mglShadeState = 1;

  mglFillState = 1;
  glPolygonMode (GL_FRONT, GL_FILL);

  mglLineState = 1.0;
  glLineWidth (mglLineState);

  mglPointState = 1.0;
  glPointSize (mglPointState);

  mglTextureEnableState = 0;
  glDisable (GL_TEXTURE_2D);

  mglBeginEndState = 0;

  mglSetRectangleColors (mglWhite);
  mglSetTriangleColors (mglWhite);
  mglSetTextureColors (mglWhite);
  mglSetLineColors (mglWhite);

  mglPointColor = mglWhite;

  mglSetColor (mglWhite);
//  mglColorMode = 0;

  mglModeState = -1;

  mglLastUsedTexture = -1;

  mglSetTextureZoom (1);

  glEnable (GL_BLEND);

  c.a = c.b = c.r = c.g = -1;
  mglSetColor (c);
}

static int _mglOn_ (int *var)
{
	if (*var == 0)
	{
		_mglAbort_ ();
		*var = 1;
		mglTestOpenGLError ();
		return 1;
	}
	return 0;
}

static int _mglOff_ (int *var)
{
	if (*var)
	{
		_mglAbort_ ();
		*var = 0;
		mglTestOpenGLError ();
		return 1;
	}
	return 0;
}

static int _mglEnable_ (int *var, int gl)
{
	if (_mglOn_ (var))
	{
		glEnable (gl);
		return 1;
	}
	return 0;
}

static int _mglDisable_ (int *var, int gl)
{
	if (_mglOff_ (var))
	{
		glDisable (gl);
		return 1;
	}
	return 0;
}

void mglPointSmoothOn (void)
{
	if (_mglEnable_ (&mglPointSmoothState, GL_POINT_SMOOTH))
		glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
}

void mglPointSmoothOff (void)
{
	_mglDisable_ (&mglPointSmoothState, GL_POINT_SMOOTH);
}

void mglSmoothOn (void)
{
	mglPointSmoothOn ();

	if (_mglOn_ (&mglSmoothState))
	{
		if (_strnicmp (mglRenderer, "RIVA", 4) != 0)
		{
			glEnable (GL_LINE_SMOOTH);
			glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
		}
		glEnable (GL_POLYGON_SMOOTH);
		glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);
		
		mglTestOpenGLError ();
	}
}

void mglSmoothOff (void)
{
	mglPointSmoothOff ();

	if (_mglOff_ (&mglSmoothState))
	{
		glDisable (GL_LINE_SMOOTH);
		glDisable (GL_POINT_SMOOTH);
		glDisable (GL_POLYGON_SMOOTH);
	}
}

void mglAlphaOn (void)
{
	_mglEnable_ (&mglAlphaState, GL_BLEND);
}

void mglAlphaOff (void)
{
	_mglDisable_ (&mglAlphaState, GL_BLEND);
}

void mglLineStippleOn (void)
{
	_mglEnable_ (&mglLineStippleState, GL_LINE_STIPPLE);
}

void mglLineStippleOff (void)
{
	_mglDisable_ (&mglLineStippleState, GL_LINE_STIPPLE);
}

void mglSetLineStipple (int factor, unsigned short pattern)
{
	_mglAbort_ ();
	glLineStipple (factor, pattern);
}

void mglInterpolationOn (void)
{
	if (_mglOn_ (&mglShadeState))
		glShadeModel (GL_SMOOTH);
}

void mglInterpolationOff (void)
{
	if (_mglOff_ (&mglShadeState))
		glShadeModel (GL_FLAT);
}

static void mglInitialWindow (void)
{
	_mglOff_ (&mglFullScreenState);

	// The X value of the initial window position. 
	// The Y value of the initial window position. 
	glutPositionWindow (GLUT_INIT_WINDOW_X, GLUT_INIT_WINDOW_Y);

	// The width value of the initial window size. 
	// The height value of the initial window size. 
	glutReshapeWindow (GLUT_INIT_WINDOW_WIDTH, GLUT_INIT_WINDOW_HEIGHT);
}



void mglFullScreenOn (void)
{
	if (_mglOn_ (&mglFullScreenState))
	{
		mglWindowSizeX = mglWindowWidth;
		mglWindowSizeY = mglWindowHeight;
		mglWindowPosX = glutGet (GLUT_WINDOW_X);
		mglWindowPosY = glutGet (GLUT_WINDOW_Y);

		glutFullScreen ();
	}
}

void mglFullScreenOff (void)
{
	if (_mglOff_ (&mglFullScreenState))
	{
		glutPositionWindow (mglWindowPosX, mglWindowPosY);
		glutReshapeWindow (mglWindowSizeX, mglWindowSizeY);
	}
}


void _mglTexturesOn_ (void)
{
	_mglEnable_ (&mglTextureEnableState, GL_TEXTURE_2D);
}

void _mglTexturesOff_ (void)
{
	_mglDisable_ (&mglTextureEnableState, GL_TEXTURE_2D);
}


float mglSetLineSize (float size)
{
  if (size > mglMaxLineSize)
    size = mglMaxLineSize;

  if (size != mglLineState)
  {
    _mglAbort_ ();
    glLineWidth (size);
    mglTestOpenGLError ();
    mglLineState = size;
  }

  return size;
}

// glLineStipple nebude podporovan

// GL_ALPHA_TEST
// GL_BLEND



float mglSetPointSize (float size)
{
  if (size > mglMaxPointSize)
    size = mglMaxPointSize;

  if (size != mglPointState)
  {
    _mglAbort_ ();
     glPointSize (size);
    mglTestOpenGLError ();
    mglPointState = size;
  }

  return size;
}




/*
void mglSetColor (float r, float g, float b, float a)
{
  glColor4f (r, g, b, a);
}
*/

void mglSetGrayColor (float rgb, float a)
{
	mglColor c;
	c.r = c.g = c.b = rgb; c.a = a;
	mglSetColor (c);
}

void mglDrawPoint (float x, float y)
{
  _mglFillOn_ ();
  _mglTexturesOff_ ();

  _mglBegin_ (GL_POINTS);
  mglSetColor (mglPointColor);
  glVertex2f (x, y);
  _mglEnd_ ();
}

void _mglRotationOn_ (float sx, float sy)
{
  _mglAbort_ ();
  glTranslatef (sx, sy, 0);
  glRotatef (mglRotationAngle, 0, 0, 1);
}

void _mglRotationOff_ (float sx, float sy)
{
  _mglAbort_ ();
  glRotatef (-mglRotationAngle, 0, 0, 1);
  glTranslatef (-sx, -sy, 0);
}


void mglDrawLine (float x, float y, float maxx, float maxy)
{
  float sx, sy;

  if (mglRotationAngle)
  {
    sx = x + (maxx - x) / 2;
    sy = y + (maxy - y) / 2;
    _mglRotationOn_ (sx, sy);
    x = x - sx;
    y = y - sy;
    maxx = -x;
    maxy = -y;
  }
  else
    sx = sy = 0;

  _mglFillOff_ ();
  _mglTexturesOff_ ();

  mglSetColor (mglLineColor [0]);
  _mglBegin_ (GL_LINES);
  glVertex2f (x, y);
  mglSetColor (mglLineColor [1]);
  glVertex2f (maxx, maxy);
  _mglEnd_ ();

  if (mglRotationAngle)
  {
    _mglRotationOff_ (sx, sy);
  }
}

void mglSetLineColors (mglColor color)
{
  mglLineColor [0] = color;
  mglLineColor [1] = color;
}


void mglSetRectangleColors (mglColor color)
{
  mglRectangleColor [0] = color;
  mglRectangleColor [1] = color;
  mglRectangleColor [2] = color;
  mglRectangleColor [3] = color;
}



void mglSetTriangleColors (mglColor color)
{
  mglTriangleColor [0] = color;
  mglTriangleColor [1] = color;
  mglTriangleColor [2] = color;
}


/*
void mglSetRectColor (mglColor color [4]);
{
  mglRectColor = color;
}
*/

/*
void mglSetTriangleColor (mglColor color [3]);
{
  mglRectangleColor [0] = color [0];
  mglRectangleColor [1] = color [1];
  mglRectangleColor [2] = color [2];
}
*/


static void mglTriangleCenter (float *sx, float *sy, float x1, float y1, float x2, float y2, float x3, float y3)
{
  int nx1, ny1, nx2, ny2, nx3, ny3;

  // ubohy algoritmus na spocteni stredu trojuhelniku ... 

  do
  {
    nx1 = x1;
    nx2 = x2;
    nx3 = x3;
    ny1 = y1;
    ny2 = y2;
    ny3 = y3;

    x1 = (nx1 + nx2) / 2;
    y1 = (ny1 + ny2) / 2;
    x2 = (nx1 + nx3) / 2;
    y2 = (ny1 + ny3) / 2;
    x3 = (nx2 + nx3) / 2;
    y3 = (ny2 + ny3) / 2;

  }
  while ((nx1 != x1) || (ny1 != y1) || (nx2 != x2) || (ny2 != y2) || (nx3 != x3) ||
         (ny3 != y3));

  *sx = nx1;
  *sy = ny2;
}

void mglTriangleEngine (float x1, float y1, float x2, float y2, float x3, float y3)
{
  float sx, sy;

  _mglTexturesOff_ ();

  if (mglRotationAngle)
  {
    mglTriangleCenter (&sx, &sy, x1, y1, x2, y2, x3, y3);
    _mglRotationOn_ (sx, sy);
    x1 -= sx; y1 -= sy; x2 -= sx; y2 -= sy; x3 -= sx; y3 -= sy;
  }

  mglSetColor (mglTriangleColor [0]);
  _mglBegin_ (GL_TRIANGLES);
  glVertex2f (x1, y1);
  mglSetColor (mglTriangleColor [1]);
  glVertex2f (x2, y2);
  mglSetColor (mglTriangleColor [2]);
  glVertex2f (x3, y3);
  _mglEnd_ ();

  if (mglRotationAngle)
  {
    _mglRotationOff_ (sx, sy);
  }
}


void mglRectangleEngine (float x, float y, float maxx, float maxy)
{
  float sx, sy;

  _mglTexturesOff_ ();

  if (mglVirtualStartX > mglVirtualEndX)
  {
	  sx = x; x = maxx; maxx = sx;
  }

  if (mglVirtualStartY > mglVirtualEndY)
  {
	  sy = y; y = maxy; maxy = sy;
  }

  if (mglRotationAngle)
  {
    sx = x + (maxx - x) / 2;
    sy = y + (maxy - y) / 2;
    _mglRotationOn_ (sx, sy);
    x = x - sx;
    y = y - sy;
    maxx = -x;
    maxy = -y;
  }
  else
    sx = sy = 0;

  mglSetColor (mglRectangleColor [0]);
  _mglBegin_ (GL_QUADS);
  glVertex2f (x, y);
  mglSetColor (mglRectangleColor [1]);
  glVertex2f (maxx, y);
  mglSetColor (mglRectangleColor [2]);
  glVertex2f (maxx, maxy);
  mglSetColor (mglRectangleColor [3]);
  glVertex2f (x, maxy);
  _mglEnd_ ();

  if (mglRotationAngle)
    _mglRotationOff_ (sx, sy);
}



void mglDrawFilledTriangle (float x1, float y1, float x2, float y2, float x3, float y3)
{
  _mglFillOn_ ();

  mglTriangleEngine (x1, y1, x2, y2, x3, y3);
}

void mglDrawTriangle (float x1, float y1, float x2, float y2, float x3, float y3)
{
  _mglFillOff_ ();

  mglTriangleEngine (x1, y1, x2, y2, x3, y3);
}




void mglDrawRectangle (float x, float y, float sizeX, float sizeY)
{
  _mglFillOff_ ();

  mglRectangleEngine (x, y, x + sizeX, y + sizeY);
}

void mglDrawFilledRectangle (float x, float y, float sizeX, float sizeY)
{
  _mglFillOn_ ();

  mglRectangleEngine (x, y, x + sizeX, y + sizeY);
}

void mglDrawFilledTriangleRectangle (float x, float y, float sx, float sy, mglColor c)
{
	float mx=x+sx/2, my = y+sy/2;

	mglTriangleColor [0] = mglRectangleColor [3];
	mglTriangleColor [1] = c;
	mglTriangleColor [2] = mglRectangleColor [2];

	mglDrawFilledTriangle (x, y, mx, my, x + sx, y);

	mglTriangleColor [1] = mglRectangleColor [0];
	mglTriangleColor [2] = c;
	mglDrawFilledTriangle (x, y, x, y + sy, mx, my);

	mglTriangleColor [0] = c;
	mglTriangleColor [2] = mglRectangleColor [1];
	mglDrawFilledTriangle (mx, my, x, y + sy, x + sx, y + sy);

	mglTriangleColor [1] = mglRectangleColor [1];
	mglTriangleColor [2] = mglRectangleColor [2];
	mglDrawFilledTriangle (mx, my, x + sx, y + sy, x + sx, y);
}


extern int mglExtTexture;

static void _mglCheckTextureResidence_ (mglTexture *texture)
{
    int size;
	GLint params;

	if (texture->residancechecked == 0)
	{
  	  _mglAbort_ ();

	  size = texture->image.w*texture->image.h*texture->image.d;

	  if (mglExtTexture)
		glGetTexParameteriv (GL_TEXTURE_2D, GL_TEXTURE_RESIDENT, &params);
	  else
		params = 0;

	  if (params)
	  {
		mglResidentTexturesTotal++;
		mglResidentTexturesMemory+=size;
		texture->residancechecked = 2;
	  }
	  else
	  {
		mglLocalTexturesTotal++;
		mglLocalTexturesMemory+=size;
		texture->residancechecked = 1;
	  }
	}
}

void mglDrawTexturedEngine (double tx1, double ty1, double tx2, double ty2,
                            mglTexture *texture, float x, float y, float maxx, float maxy)
{
  int xr;
  int yr;
  float sx;
  float sy;

  if (mglVirtualStartX > mglVirtualEndX)
  {
	  sx = x; x = maxx; maxx = sx;
  }

  if (mglVirtualStartY > mglVirtualEndY)
  {
	  sy = y; y = maxy; maxy = sy;
  }

//  _mglAbort_ ();
//  mglFillState = 0;
//  _mglFillOn_ ();

  xr = texture->image.xrev; yr = texture->image.yrev;

  if (mglRotationAngle)
  {
    sx = x + (maxx - x) / 2;
    sy = y + (maxy - y) / 2;
    _mglRotationOn_ (sx, sy);
    x = x - sx;
    y = y - sy;
    maxx = -x;
    maxy = -y;
  }
  else
    sx = sy = 0;

  _mglFillOn_ ();

  _mglTexturesOn_ ();

  mglChangeTexture (texture);

  _mglBegin_(GL_QUADS);

  mglSetColor (mglTextureColor [0]);
  glTexCoord2f(xr ? tx2 : tx1, yr ? ty1 : ty2); 
  glVertex2f(x, y);

  mglSetColor (mglTextureColor [1]);
  glTexCoord2f(xr ? tx1 : tx2, yr ? ty1 : ty2); 
  glVertex2f(maxx, y);

  mglSetColor (mglTextureColor [2]);
  glTexCoord2f(xr ? tx1 : tx2, yr ? ty2 : ty1); 
  glVertex2f(maxx, maxy);

  mglSetColor (mglTextureColor [3]);
  glTexCoord2f(xr ? tx2 : tx1, yr ? ty2 : ty1); 
  glVertex2f(x, maxy);

  _mglEnd_();

  if (mglRotationAngle)
  {
    _mglRotationOff_ (sx, sy);
  }

  _mglCheckTextureResidence_ (texture);
}

void mglDrawTexturedEngine4 (double tx1, double ty1, double tx2, double ty2, double tx3, double ty3, double tx4, double ty4,
                            mglTexture *texture, float x, float y, float maxx, float maxy)
{
  int xr;
  int yr;
  float sx;
  float sy;

  if (mglVirtualStartX > mglVirtualEndX)
  {
	  sx = x; x = maxx; maxx = sx;
  }

  if (mglVirtualStartY > mglVirtualEndY)
  {
	  sy = y; y = maxy; maxy = sy;
  }

//  _mglAbort_ ();
//  mglFillState = 0;
//  _mglFillOn_ ();

  xr = texture->image.xrev; yr = texture->image.yrev;

  if (mglRotationAngle)
  {
    sx = x + (maxx - x) / 2;
    sy = y + (maxy - y) / 2;
    _mglRotationOn_ (sx, sy);
    x = x - sx;
    y = y - sy;
    maxx = -x;
    maxy = -y;
  }
  else
    sx = sy = 0;

  _mglFillOn_ ();

  _mglTexturesOn_ ();

  mglChangeTexture (texture);

  _mglBegin_(GL_QUADS);

  mglSetColor (mglTextureColor [0]);
  glTexCoord2f(xr ? tx2 : tx1, yr ? ty1 : ty4); 
  glVertex2f(x, y);

  mglSetColor (mglTextureColor [1]);
  glTexCoord2f(xr ? tx1 : tx2, yr ? ty2 : ty3); 
  glVertex2f(maxx, y);

  mglSetColor (mglTextureColor [2]);
  glTexCoord2f(xr ? tx4 : tx3, yr ? ty2 : ty3); 
  glVertex2f(maxx, maxy);

  mglSetColor (mglTextureColor [3]);
  glTexCoord2f(xr ? tx3 : tx4, yr ? ty4 : ty1); 
  glVertex2f(x, maxy);

  _mglEnd_();

  if (mglRotationAngle)
  {
    _mglRotationOff_ (sx, sy);
  }

  _mglCheckTextureResidence_ (texture);
}

extern int mglWinWidth;
extern int mglWinHeight;
extern int mglRegionX;
extern int mglRegionY;
extern int mglRegionHeight;
extern int mglRegionWidth;

void mglClearScreen (mglColor clearColor)
{
  // setups background color to black
	int i = 1;

  clearColor = mglAdjustColor (clearColor, mglColorMode);

  glClearColor (clearColor.r, clearColor.g, clearColor.b, clearColor.a);

  glClear (GL_COLOR_BUFFER_BIT /*| GL_DEPTH_BUFFER_BIT*/);

}

