
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include <mgl.h>

int mglWindowWidth;
int mglWindowHeight;
int mglReshaped;

int mglWaitForReshape;

static mglHandlerType mglDisplay;
static mglHandlerType mglIdle;

int mglRegionX;
int mglRegionY;

int mglScreenWidth;
int mglScreenHeight;

static int mglInitialized;

extern void mglInitState (void);
extern void mglDoneState (void);

int mglInitGLUT (void)
{
  char *av [] = {""};
  int ac = 1;


  if (!mglInitialized)
  {
    mglInitialized = 1;

    // nainicializuje knihovnu glut s prazdnymi prepinaci
    glutInit (&ac, av);

    // nastavi rezim zobrazovani
    // GLUT_DOUBLE pro umozneni plynule animace a jinych efektu. Rezim zobrazovani je RGB
    glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);

    mglScreenWidth = glutGet (GLUT_SCREEN_WIDTH);
    mglScreenHeight = glutGet (GLUT_SCREEN_HEIGHT);
  }

  return 0;
}

extern void mglTimer (void);

static void mglDisplayFunc (void)
{
  extern int mglLastFrames;

  mglTimer ();

  if (mglWindowWidth)
  {
    if (!mglWaitForReshape)
    {
      mglInitState ();

      if (mglDisplay)
        mglDisplay ();

      mglDoneState ();

      glFinish ();

      glutSwapBuffers ();
      mglFrames++;
      mglLastFrames++;

      if (!mglWaitForReshape)
        mglReshaped = 0;
    }
  }
}


static void mglIdleFunc (void)
{
  //mglDisplayFunc ();
  if (mglIdle)
    mglIdle ();
}


void mglSetDisplayHandler (mglHandlerType handler)
{
  mglDisplay = handler;
}


void mglSetIdleHandler (mglHandlerType handler)
{
  mglIdle = handler;
}

extern void mglSetupMGLMode (void);

int mglOpenWindow (char *title, int x, int y, int sizeX, int sizeY)
{
  if (!mglVirtualWidth)
    mglVirtualWidth = sizeX;
  if (!mglVirtualHeight)
    mglVirtualHeight = sizeY;

  mglNote ("\n---------------------------\n");
  mglNote ("Initialiazing GLUT ...\n");
  mglInitGLUT ();

  // nastaveni pocatecni pozici a velikost okna
  glutInitWindowSize(sizeX, sizeY);
  glutInitWindowPosition(x, y);

  // vytvori okno s nazvem aplikace
  mglNote ("Creating window ...\n");
  glutCreateWindow (title);
  mglWaitForReshape = 1;

  // tato funkce musi byt zavolana az po vytvoreni okna
//  mglGetGLInfo ();
  mglSetupMGLMode ();
  mglNote ("Window ready.\n");

  return 0;
}

int mglFrames;
int mglLastFrames;


int mglOpenFullscreen (char *title)
{
  mglOpenWindow (title, 100, 100, 320, 240);

  mglFullScreenOn ();

  return 0;
}


// funkce, ktere volaji opengl k obnoveni obsahu oken

static int newVirtualWidth;
static int newVirtualHeight;

int mglWinWidth;
int mglWinHeight;
int mglRegionX;
int mglRegionY;
int mglRegionHeight;
int mglRegionWidth;

extern int mglFullScreenState;

static void mglReshapeFunc (int width, int height)
{
  int w;
  int h;
  int regionx;
  int regiony;

  if ((mglWinWidth != width) || (mglWinHeight != height) || (newVirtualWidth != mglVirtualWidth) || (newVirtualHeight != mglVirtualHeight))
  {
    mglWinWidth = width;
    mglWinHeight = height;
    newVirtualWidth = mglVirtualWidth;
    newVirtualHeight = mglVirtualHeight;
  }
  else
    return;

// vycisti buffer pristiho ramce

  regionx = regiony = 0;

  w = height * mglVirtualWidth / mglVirtualHeight;
  h = width  * mglVirtualHeight / mglVirtualWidth;

  if (width > w)
  {
    regionx = (width - w)/2;
    width = w;
  }
  if (height > h)
  {
    regiony = (height - h)/2;
    height = h;
  }

  mglReshaped = 1;
  mglLastFrames = 0;
  mglTimer ();

  mglRegionX = regionx;
  mglRegionY = regiony;

  glViewport(regionx, regiony, width, height); /* viewport size in pixels */
  mglRegionWidth = width;
  mglRegionHeight = height;

  mglWindowWidth = width;
  mglWindowHeight = height;

  mglSetupMGLMode ();

  mglWaitForReshape = 0;
}

extern void mglInitInput (void);

void mglRun (void)
{
  glutReshapeFunc(mglReshapeFunc);

  // deklarace uzivatelskych funkci
  glutDisplayFunc(mglDisplayFunc);
  glutIdleFunc (mglDisplayFunc);
  mglInitInput ();
  mglFrames = 0;
  mglLastFrames = 0;

  glutMainLoop();
}

static char oldTitle [333];

void mglSetWindowTitle (char *title)
{
  if (strcmp (oldTitle, title))
  {
    mglDoneState ();
    glutSetWindowTitle (title);
    strcpy (oldTitle, title);
  }
}

void mglSetWindowSize (int x, int y)
{
  x = (x / 8) * 8;
  y = (y / 8) * 8;
  if (!x)
    x = 8;
  if (!y)
    y = 8;

  mglDoneState ();

  if ((mglWinWidth == x) && (mglWinHeight == y))
  {
    mglReshapeFunc (x, y);
    return;
  }

  glutReshapeWindow (x, y);
  mglWaitForReshape = 1;
}

